import asyncio
import unittest
from typing import *

import outleap


class MockLEAPProtocol(outleap.AbstractLEAPProtocol):
    def __init__(self):
        self.closed = False
        self.sent_messages = []
        self.inbound_messages = asyncio.Queue()

    def close(self) -> None:
        self.closed = True

    def write_message(self, pump: str, data: Any) -> None:
        self.sent_messages.append({"pump": pump, "data": data})

    async def read_message(self) -> Dict:
        msg = await self.inbound_messages.get()
        self.inbound_messages.task_done()
        return msg


class MockReqIDGenerator:
    """Predictable reqid generator for request reply tracking"""

    def __init__(self):
        self.ctr = 0

    def __call__(self, *args, **kwargs):
        self.ctr += 1
        return self.ctr


class BaseClientTest(unittest.IsolatedAsyncioTestCase):
    async def asyncSetUp(self) -> None:
        self.loop = asyncio.get_running_loop()
        self.protocol = MockLEAPProtocol()
        self.client = outleap.LEAPClient(self.protocol)
        self.client._gen_reqid = MockReqIDGenerator()

    async def asyncTearDown(self) -> None:
        self.client.disconnect()

    def _write_welcome(self):
        self.protocol.inbound_messages.put_nowait(
            {
                "pump": "reply_pump",
                "data": {
                    "command": "cmd_pump",
                },
            }
        )

    def _write_reply(self, reqid: int, extra: Optional[Dict] = None):
        self.protocol.inbound_messages.put_nowait(
            {
                "pump": "reply_pump",
                "data": {
                    **(extra or {}),
                    "reqid": reqid,
                },
            }
        )



from __future__ import annotations

import asyncio
import collections
import dataclasses
import weakref
from typing import *

if TYPE_CHECKING:
    from .api_wrappers import LLWindowAPI


class UIPath:
    __slots__ = ["_parts", "__weakref__"]
    # UIPaths are immutable, so we can maintain a cache of instances based on
    # the parts that make up the path.
    _INSTANCES: Dict[Tuple[str, ...], UIPath] = weakref.WeakValueDictionary()

    def __new__(cls, *args):
        if len(args) == 1:
            if isinstance(args[0], str):
                args = args[0].split("/")
                # "." means "/" in LEAP
                if tuple(args) == (".",):
                    args = ()
                else:
                    # Ignore empty segments
                    args = tuple([x for x in args if x])
            elif isinstance(args[0], UIPath):
                return args[0]
            else:
                args = tuple(args)
        else:
            args = tuple(args)

        # Check if we have a cached UIPath instance for these args first
        val = cls._INSTANCES.get(args)
        if val is not None:
            return val
        val = super().__new__(cls)
        val._parts = args
        cls._INSTANCES[args] = val
        return val

    @classmethod
    def for_floater(cls, floater_name: str) -> UIPath:
        return cls("/main_view/menu_stack/world_panel/Floater View") / floater_name

    def __str__(self) -> str:
        return "/" + "/".join(self._parts)

    def __repr__(self):
        return f"{self.__class__.__name__}({self!s})"

    def __eq__(self, other):
        if isinstance(other, UIPath):
            return self is other
        elif isinstance(other, str):
            return UIPath(other) == self
        return NotImplemented

    def __hash__(self):
        return hash(self._parts)

    @property
    def parent(self) -> UIPath:
        return UIPath(*self._parts[:-1])

    @property
    def stem(self) -> str:
        if self._parts:
            return self._parts[-1]
        return ""

    def __truediv__(self, other: Union[str, UIPath]):
        if not isinstance(other, (UIPath, str)):
            return NotImplemented
        if isinstance(other, str):
            other = UIPath(other)
        return UIPath(*self._parts, *other._parts)

    def __rtruediv__(self, other: Union[str, UIPath]):
        if not isinstance(other, (UIPath, str)):
            return NotImplemented
        if isinstance(other, str):
            other = UIPath(other)
        return UIPath(*other._parts, *self._parts)

    def is_relative_to(self, other: Union[str, UIPath]) -> bool:
        # 10x faster than pathlib's `is_relative_to()`.
        if isinstance(other, str):
            other = UIPath(other)
        other_len = len(other._parts)
        self_len = len(self._parts)
        # self should be a strict extension of other
        return self_len >= other_len and self._parts[:other_len] == other._parts


class UIRect(NamedTuple):
    bottom: int
    left: int
    right: int
    top: int


@dataclasses.dataclass
class UIElementInfo:
    available: bool
    class_name: str
    enabled: bool
    enabled_chain: bool
    rect: UIRect
    value: Any
    visible: bool
    visible_chain: bool

    def to_dict(self):
        return {
            "available": self.available,
            "class": self.class_name,
            "enabled": self.enabled,
            "enabled_chain": self.enabled_chain,
            "rect": self.rect._asdict(),
            "value": self.value,
            "visible": self.visible,
            "visible_chain": self.visible_chain,
        }


_T = TypeVar("T")


class UIElementProperty(Generic[_T]):
    __slots__ = ("name", "owner")

    def __init__(self):
        super().__init__()

    def __set_name__(self, owner, name: str):
        self.name = name

    def __get__(self, _obj: UIElement, owner: Optional[Type] = None) -> _T:
        # Reach into the UIElementTree for info related to the element our owner wraps.
        # Try to access the parameter related to this. This will intentionally fail if
        # the UIElement info aren't complete.
        return getattr(_obj.info, self.name)


class UIElement:
    """Wrapper around a tree and a path to access that path in the tree"""

    def __init__(self, path: UIPath, element_tree: UIElementTree):
        self.path = path
        self._tree = element_tree

    def __str__(self) -> str:
        return str(self.path)

    def __repr__(self):
        return f"{self.__class__.__name__}({str(self)!r})"

    @property
    def info(self) -> Optional[UIElementInfo]:
        return self._tree.get_info_by_path(self.path)

    @property
    def is_detached(self) -> bool:
        """Whether this element is still a valid node in its parent tree"""
        try:
            self._tree.get_info_by_path(self.path)
            return False
        except KeyError:
            return True

    @property
    def parent(self) -> Optional[UIElement]:
        parent = self.path.parent
        if parent == self.path:
            return None
        return self._tree[parent]

    @property
    def ancestors(self) -> Sequence[UIElement]:
        ancestors = []
        elem = self
        while elem := elem.parent:
            ancestors.append(elem)
        return ancestors

    def to_dict(self):
        info = self.info
        return {
            "path": self.path,
            **(info.to_dict() if info else {}),
        }

    @property
    def children(self) -> List[UIElement]:
        return self._tree.get_child_elems(self.path)

    available: bool = UIElementProperty()
    """Visible and may be interacted with"""
    class_name: str = UIElementProperty()
    """C++ class name of the element"""
    enabled: bool = UIElementProperty()
    """Whether this specific element is enabled"""
    enabled_chain: bool = UIElementProperty()
    """Whether everything in the element's ancestry is enabled"""
    rect: UIRect = UIElementProperty()
    """Extents of the element rect, not clipped to parent rects"""
    value: Any = UIElementProperty()
    """Logical value of the element, LLSD"""
    visible: bool = UIElementProperty()
    """Whether this specific element is set to be visible"""
    visible_chain: bool = UIElementProperty()
    """Whether everything in the element's ancestry is visible"""

    def __eq__(self, other) -> bool:
        if not isinstance(other, UIElement):
            return False
        if other._tree != self._tree:
            return False
        return other.path == self.path

    async def refresh_subtree(self, refresh_info=False):
        await self._tree.refresh_subtree(self.path, refresh_info=refresh_info)

    async def refresh(self):
        await self._tree.fetch_info_for_paths([self.path])


class UIElementTree(Mapping[UIPath, "UIElement"]):
    def __init__(self, window_api: LLWindowAPI):
        self._window_api: LLWindowAPI = window_api
        self._elem_info: Dict[UIPath, Optional[UIElementInfo]] = {}
        self._children: Dict[UIPath, List[UIPath]] = collections.defaultdict(list)
        super().__init__()

    def __getitem__(self, key: Union[str, UIPath]) -> UIElement:
        if isinstance(key, str):
            key = UIPath(key)
        if key not in self._elem_info:
            raise KeyError(f"No element at {key}")
        return UIElement(key, self)

    def __contains__(self, item: Union[str, UIPath]) -> bool:
        if isinstance(item, str):
            item = UIPath(item)
        return item in self._elem_info

    def __len__(self) -> int:
        return len(self._elem_info)

    def __iter__(self) -> Iterator[UIPath]:
        return iter(self._elem_info)

    @property
    def root_children(self) -> List[UIElement]:
        return self.get_child_elems(UIPath("/"))

    def get_info_by_path(self, path: UIPath) -> UIElementInfo:
        return self._elem_info[path]

    def get_child_elems(self, path: UIPath) -> List[UIElement]:
        return [UIElement(x, self) for x in self._children.get(path, ())]

    async def refresh_subtree(self, under: Optional[UIPath] = None, refresh_info: bool = False):
        """
        Drop and refresh all info about elements under `path`

        If refresh_info is specified, fetch additional info about all elements
        found, otherwise just use existing info about the elements if available.
        """
        # Elements may have been dropped, get rid of all elements under this path.
        old_paths: Dict[UIPath, Optional[UIElementInfo]] = {}
        for key, val in list(self._elem_info.items()):
            if not under or key.is_relative_to(under):
                old_paths[key] = self._elem_info[key]
                del self._elem_info[key]
                self._children.pop(key, None)

        new_paths = await self._window_api.get_paths(under=under)

        for path in new_paths:
            # Fill map with either the old data or "None". Presence of "None"
            # distinguishes paths which don't exist within the tree from ones
            # which do exist, but for which we have no data.
            self._elem_info[path] = old_paths.get(path)
            if path.parent != path:
                child_elems = self._children[path.parent]
                # We want order of elements, but we also might get dupes, which
                # we don't want...
                if path not in child_elems:
                    child_elems.append(path)

        if refresh_info:
            await self.fetch_info_for_paths(new_paths)

    async def fetch_info_for_paths(self, paths: Collection[UIPath], allow_missing: bool = True):
        futs = [self._window_api.get_info(path) for path in paths]
        for path, info in zip(paths, await asyncio.gather(*futs)):
            if error := info.get("error"):
                if "request specified invalid" in error and allow_missing:
                    # Implicitly null elem_info for invalid paths. There are certain paths that
                    # are sort of valid but that the viewer refuses to respond to getInfo for.
                    # Don't remove them from the path tree, just say we don't have info.
                    elem_info = None
                else:
                    raise ValueError(error)
            else:
                elem_info = UIElementInfo(
                    available=info["available"],
                    class_name=info["class"],
                    # Some things are represented in the LLSD payload as int even though
                    # they'd be better represented as `bool`s. This is just an artifact
                    # of the functions in the C++ side using `BOOL` rather than `bool`.
                    enabled=bool(info["enabled"]),
                    enabled_chain=bool(info["enabled_chain"]),
                    rect=UIRect(**info["rect"]),
                    value=info.get("value"),
                    visible=bool(info["visible"]),
                    visible_chain=bool(info["visible_chain"]),
                )

            self._elem_info[UIPath(path)] = elem_info

    async def refresh(self, refresh_info: bool = False):
        await self.refresh_subtree(under=None, refresh_info=refresh_info)


UI_PATH_TYPE = Optional[Union[str, UIPath, UIElement]]


__all__ = [
    "UIPath",
    "UIRect",
    "UIElement",
    "UIElementTree",
    "UI_PATH_TYPE",
]



"""
Interactive REPL that handles connect-back connections from outleap-agent
"""
import asyncio
import logging
import multiprocessing
import pprint
import sys
import uuid

try:
    import ptpython
except ImportError:
    print("ptpython must be installed to use the REPL", file=sys.stderr)
    raise

import outleap


class REPLServer:
    def __init__(self):
        self._repl_running = False

    async def client_connected(self, client: outleap.LEAPClient):
        if self._repl_running:
            logging.error("Client already connected, ignoring incoming connection")
            return

        self._repl_running = True

        new_globals = {
            **globals(),
            "pprint": pprint,
            "UUID": uuid.UUID,
        }
        # Simulate `from outleap import *` in the REPL's global environment
        for name in outleap.__all__:
            new_globals[name] = getattr(outleap, name)

        try:
            await ptpython.repl.embed(  # noqa: the type signature lies
                globals=new_globals,
                locals={"client": client},
                return_asyncio_coroutine=True,
                patch_stdout=False,
            )
        finally:
            self._repl_running = False


def repl_main():
    logging.basicConfig()
    loop = asyncio.get_event_loop_policy().get_event_loop()
    repl_server = REPLServer()
    server = outleap.LEAPBridgeServer(repl_server.client_connected)
    coro = asyncio.start_server(server.handle_connection, "127.0.0.1", 9063)
    loop.run_until_complete(coro)
    print("REPL listening for inbound outleap-agent connections!", file=sys.stderr)
    loop.run_forever()


if __name__ == "__main__":
    multiprocessing.freeze_support()
    repl_main()



"""
Code related to the stdin/stdout -> TCP connectback bridge
"""

from __future__ import annotations

import asyncio
import logging
import weakref
from typing import *

from .client import LEAPClient
from .protocol import LEAPProtocol


class LEAPBridgeServer:
    """LEAP Bridge TCP server to use with asyncio.start_server()"""

    def __init__(self, client_connected_cb: Optional[Callable[[LEAPClient]], Awaitable[Any]] = None):
        self.clients: weakref.WeakSet[LEAPClient] = weakref.WeakSet()
        self._client_connected_cb = client_connected_cb

    async def handle_connection(self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter):
        client = LEAPClient(LEAPProtocol(reader, writer))
        logging.info("Accepting LEAP connection from %r" % (writer.get_extra_info("peername", None),))
        await client.connect()

        self.clients.add(client)
        if self._client_connected_cb:
            await self._client_connected_cb(client)


__all__ = [
    "LEAPBridgeServer",
]



import asyncio
import pathlib
import unittest
from typing import *

import outleap
from outleap import UIPath

from . import MockLEAPProtocol

THIS_DIR = pathlib.Path(__file__).parent.absolute()
UI_PATHS: List[str] = []
with open(THIS_DIR / "ui_paths.txt", "r") as f:
    for path in f:
        UI_PATHS.append(path.strip())


class MockLEAPClient(outleap.LEAPClient):
    def command(self, pump: str, op: str, data: Optional[Dict] = None) -> Optional[asyncio.Future]:
        fut = asyncio.Future()
        if pump == "LLWindow":
            if op == "getPaths":
                fut.set_result({"paths": UI_PATHS[:] * 20})
            else:
                assert False
        else:
            assert False
        return fut


class TestUIPaths(unittest.TestCase):
    def test_relative_to(self):
        self.assertTrue(UIPath("/foo/bar").is_relative_to(UIPath("/foo")))
        self.assertTrue(UIPath("/foo/bar").is_relative_to(UIPath("/")))
        self.assertTrue(UIPath("/").is_relative_to(UIPath("/")))
        self.assertFalse(UIPath("/").is_relative_to(UIPath("/foo")))

    def test_stem(self):
        self.assertEqual("foo", UIPath("/bar/foo").stem)
        self.assertEqual("", UIPath("/").stem)

    def test_normalize_dot(self):
        self.assertEqual("/", str(UIPath(".")))

    def test_div(self):
        self.assertEqual("/foo/bar/baz", str(UIPath("/foo/bar") / "baz"))
        self.assertEqual("/foo/bar/baz", str("/foo/bar" / UIPath("baz")))

    def test_eq(self):
        self.assertEqual(UIPath("/foo/bar"), UIPath("/foo/bar"))
        self.assertNotEqual(UIPath("/foo/bar"), UIPath("/foo"))
        self.assertEqual(UIPath("/foo/bar"), "/foo/bar")

    def test_get_parent(self):
        self.assertEqual(UIPath("/"), UIPath("/foo").parent)
        self.assertEqual(UIPath("/foo"), UIPath("/foo/bar").parent)


class TestUIElems(unittest.IsolatedAsyncioTestCase):
    async def asyncSetUp(self) -> None:
        await super().asyncSetUp()
        self.client = MockLEAPClient(MockLEAPProtocol())
        self.window_api = outleap.LLWindowAPI(self.client)
        self.tree = outleap.UIElementTree(self.window_api)

    async def test_load_paths(self):
        await self.tree.refresh()
        root_child_paths = [x.path for x in self.tree.root_children]
        self.assertListEqual(["/main_view", "/console"], root_child_paths)

    async def test_walk_children(self):
        await self.tree.refresh()

        children_map = {}

        def _walk_children(node: outleap.UIElement, parent: Optional[outleap.UIPath]):
            children_map[parent] = node
            for child in node.children:
                _walk_children(child, node.path)

        for node in self.tree.root_children:
            _walk_children(node, None)

    async def test_get_parent(self):
        await self.tree.refresh()
        self.assertEqual("/", self.tree["/main_view"].parent.path)
        self.assertIsNone(self.tree["/"].parent)



from __future__ import annotations

import asyncio
import collections
import dataclasses
import enum
import logging
import uuid
from types import TracebackType
from typing import *

from .protocol import AbstractLEAPProtocol, LEAPProtocol
from .utils import connect_stdin_stdout


class CommandPumpToken:
    pass


COMMAND_PUMP = CommandPumpToken()

PUMP_NAME_TYPE = Union[CommandPumpToken, str]


class LEAPClient:
    """Client for script -> viewer communication over the LEAP protocol"""

    def __init__(self, protocol: AbstractLEAPProtocol):
        self._protocol = protocol
        # Pump used for receiving replies
        self._reply_pump: Optional[str] = None
        # Pump used for sending leap meta-commands to the viewer (getAPIs, etc.)
        self.cmd_pump: Optional[str] = None
        # Map of req id -> future held by requester to send responses to
        self._reply_futs: Dict[uuid.UUID, asyncio.Future] = {}
        self._pump_listeners: Dict[str, ListenerDetails] = collections.defaultdict(ListenerDetails)
        self._connection_status = ConnectionStatus.READY
        self._msg_pump_task: Optional[asyncio.Task] = None
        self.shutdown_event = asyncio.Event()

    @classmethod
    async def create_stdio_client(cls) -> LEAPClient:
        """Return an already-connected LEAPClient that talks over stdin/out"""
        reader, writer = await connect_stdin_stdout()
        client = cls(LEAPProtocol(reader, writer))
        await client.connect()
        return client

    async def __aenter__(self) -> LEAPClient:
        if not self.connected:
            await self.connect()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        # Disconnect if we weren't already disconnected.
        self.disconnect()

    @property
    def connected(self) -> bool:
        return self._connection_status == ConnectionStatus.CONNECTED

    async def connect(self) -> None:
        """Receive the "hello" message from the viewer and start the message pump"""
        assert self._connection_status == ConnectionStatus.READY
        self._connection_status = ConnectionStatus.CONNECTING

        try:
            welcome_message = await self._protocol.read_message()
            self._reply_pump = welcome_message["pump"]
            self.cmd_pump = welcome_message["data"]["command"]

            self._connection_status = ConnectionStatus.CONNECTED
            self._start_message_pump()
        except:
            self.disconnect()
            raise

    def _start_message_pump(self) -> None:
        """Read and handle inbound messages in a background task"""

        async def _pump_messages_forever():
            try:
                while not self._protocol.closed:
                    self.handle_message(await self._protocol.read_message())
            except asyncio.IncompleteReadError:
                pass
            finally:
                self.disconnect()

        # Should naturally stop on its own when disconnect is called by virtue of
        # the incomplete read.
        self._msg_pump_task = asyncio.get_event_loop().create_task(_pump_messages_forever())

    def disconnect(self) -> None:
        """Close the connection and clean up any pending request futures"""
        if self.connected:
            logging.info("closing LEAP connection")
        if not self.shutdown_event.is_set():
            self.shutdown_event.set()
        self._connection_status = ConnectionStatus.DISCONNECTED
        self._protocol.close()
        if self._msg_pump_task:
            self._msg_pump_task.cancel()
            self._msg_pump_task = None

        # Clean up any pending request futures
        for fut in list(self._reply_futs.values()):
            if not fut.done():
                fut.cancel()
        self._reply_futs.clear()
        # Tell anything listening on these queues that the queues are now closed.
        for listener_details in self._pump_listeners.values():
            for listener in listener_details.listeners:
                listener.close_queue()
        self._pump_listeners.clear()

    def command(
        self, pump: PUMP_NAME_TYPE, op: str, data: Optional[Dict] = None, op_key: str = "op"
    ) -> Optional[asyncio.Future]:
        """Make a request to an internal LEAP method using the standard command form (op in data)"""
        data = data.copy() if data else {}
        data[op_key] = op
        return self.post(pump, data, expect_reply=True)

    def void_command(
        self, pump: PUMP_NAME_TYPE, op: str, data: Optional[Dict] = None, op_key: str = "op"
    ) -> None:
        """Like `command()`, but we don't expect a reply."""
        data = data.copy() if data else {}
        data[op_key] = op
        self.post(pump, data, expect_reply=False)

    def post(self, pump: PUMP_NAME_TYPE, data: Any, expect_reply: bool) -> Optional[asyncio.Future]:
        """
        Post an event to the other side's `pump`.

        Post the event is done synchronously, only waiting for the reply is done async.
        """
        assert self.connected
        if isinstance(pump, CommandPumpToken):
            pump = self.cmd_pump
        fut = None
        # If we expect a reply to this event, we need to do some extra bookkeeping.
        # There are apparently some commands for which we can never expect to get a reply.
        # Don't add a reqid or reply fut map entry in that case, since it will never be resolved.
        if expect_reply:
            # If you don't pass in a dict for data, we have nowhere to stuff `reqid`.
            # That means no reply tracking, meaning no future.
            if not isinstance(data, dict):
                raise ValueError(f"Must send a dict in `data` if you want a reply, you sent {data!r}")
            # We need to mutate the dict, make a copy so that we don't mess with the caller's version.
            data = data.copy()
            # Tell the viewer the pump to send replies to
            data["reply"] = self._reply_pump

            req_id = self._gen_reqid()
            data["reqid"] = req_id

            fut = asyncio.Future()
            # The future will be cleaned up when the Future is done.
            fut.add_done_callback(self._cleanup_request_future)
            self._reply_futs[req_id] = fut

        self._protocol.write_message(pump, data)
        return fut

    def _gen_reqid(self) -> Any:
        return uuid.uuid4()

    def listen_scoped(self, source_pump: str):
        return LEAPListenContextManager(self, source_pump)

    async def listen(self, source_pump: PUMP_NAME_TYPE) -> LEAPListener:
        """Start listening to `source_pump`, placing its messages in the returned asyncio Queue"""
        assert self.connected
        if isinstance(source_pump, CommandPumpToken):
            source_pump = self.cmd_pump

        listener_details = self._pump_listeners[source_pump]
        had_listeners = bool(listener_details.listeners)
        listener = LEAPListener()
        listener_details.listeners.add(listener)

        if not had_listeners:
            # Nothing was listening to this before, need to ask for its events to be
            # sent over LEAP.
            await self.command(
                COMMAND_PUMP,
                "listen",
                {
                    "listener": listener_details.name,
                    "source": source_pump,
                },
            )
        return listener

    async def stop_listening(self, listener: LEAPListener) -> None:
        """Stop sending a pump's messages to msg_queue, potentially removing the listen on the pump"""
        for source_pump, listener_details in self._pump_listeners.items():
            listeners = listener_details.listeners
            if listener in listeners:
                listeners.remove(listener)
                listener.close_queue()
                if self.connected and not listeners:
                    # Nobody cares about these events anymore, ask LEAP to stop sending them
                    await self.command(
                        COMMAND_PUMP,
                        "stoplistening",
                        {
                            "listener": listener_details.name,
                            "source": source_pump,
                        },
                    )
                return
        raise KeyError(f"Couldn't find {listener!r} in pump listeners")

    def handle_message(self, message: Any) -> bool:
        """Handle an inbound message and try to route it to the right recipient"""
        if not isinstance(message, dict):
            logging.warning(f"Received a non-map message: {message!r}")
            return False

        pump = message.get("pump")
        data = message.get("data")
        if pump == self._reply_pump:
            # This is a reply for a request
            if not isinstance(data, dict):
                logging.warning(f"Received a non-map reply over the reply pump: {message!r}")
                return False

            # reqid can tell us what future needs to be resolved, if any.
            fut = self._reply_futs.get(data.get("reqid"))
            if not fut:
                logging.warning(f"Received a reply over the reply pump with no reqid or future: {message!r}")
                return False
            # We don't actually care about the reqid, pop it off
            data.pop("reqid")
            # Notify anyone awaiting the response
            fut.set_result(data)
            return True

        # Might be related to a listener we registered
        # Don't warn if we get a message with an empty listener_details.queues because
        # We may still be receiving messages from before we stopped listening
        # The main concerning case if is we receive a message for something we _never_
        # registered a listener for.
        elif (listener_details := self._pump_listeners.get(pump)) is not None:
            for listener in listener_details.listeners:
                listener.put_nowait(data)
            return True
        else:
            logging.warning(f"Received a message for unknown pump: {message!r}")
        return False

    def _cleanup_request_future(self, req_fut: asyncio.Future) -> None:
        """Remove a completed future from the reply map"""
        for key, value in self._reply_futs.items():
            if value == req_fut:
                del self._reply_futs[key]
                return


class LEAPListenContextManager(AsyncContextManager[Callable[[], Awaitable[Any]]]):
    """Helper for registering and unregistering a listener within a specific scope"""

    def __init__(self, client: LEAPClient, source_pump: str):
        self._client = client
        self._source_pump = source_pump
        self._listener: Optional[LEAPListener] = None

    async def __aenter__(self) -> LEAPListener:
        self._listener = await self._client.listen(self._source_pump)
        return self._listener

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        try:
            await self._client.stop_listening(self._listener)
        except KeyError:
            pass


class ConnectionStatus(enum.Enum):
    READY = enum.auto()
    CONNECTING = enum.auto()
    CONNECTED = enum.auto()
    DISCONNECTED = enum.auto()


class LEAPListener:
    """Wrapper for queue.get() that cancels if the client disconnects while `await`ing"""

    def __init__(self):
        self._queue = asyncio.Queue()
        self._shutdown_event = asyncio.Event()

    def close_queue(self):
        if not self._shutdown_event.is_set():
            self._shutdown_event.set()

    async def get(self) -> Any:
        # Don't yield if we already have a message ready
        if not self._queue.empty():
            msg = self._queue.get_nowait()
            self._queue.task_done()
            return msg

        if self._shutdown_event.is_set():
            raise asyncio.QueueEmpty("Listener is closed and has no queued messages")

        # Wait for a queue entry to be ready, or for client shutdown
        queue_fut = asyncio.create_task(self._queue.get())
        shutdown_fut = asyncio.create_task(self._shutdown_event.wait())
        done, pending = await asyncio.wait([shutdown_fut, queue_fut], return_when=asyncio.FIRST_COMPLETED)
        if done != {queue_fut}:
            # Shutdown happened before the queue got populated
            queue_fut.cancel()
            raise asyncio.CancelledError("Client disconnected while waiting for event")
        # We found a queue entry, don't care if shutdown happens anymore.
        shutdown_fut.cancel()
        # Consumption is completion for these queues.
        self._queue.task_done()
        # Re-awaiting is fine, it'll return the previous result.
        return await queue_fut

    def empty(self) -> bool:
        return self._queue.empty()

    def put_nowait(self, val: Any) -> None:
        assert not self._shutdown_event.is_set()
        self._queue.put_nowait(val)


@dataclasses.dataclass
class ListenerDetails:
    # We can only have one listener with a given name active at a time. Give each listener a unique name.
    name: Optional[str] = dataclasses.field(default_factory=lambda: "PythonListener-%s" % uuid.uuid4())
    listeners: Set[LEAPListener] = dataclasses.field(default_factory=set)


__all__ = [
    "COMMAND_PUMP",
    "CommandPumpToken",
    "PUMP_NAME_TYPE",
    "LEAPListenContextManager",
    "LEAPListener",
    "LEAPClient",
]



"""
Tooling for working with the SL viewer's LEAP integration: now with 100% less eventlet

TODO: Support playback of VITA event recording format snippets? Does anyone actually use those?
"""

from .api_wrappers import *  # noqa
from .bridge import *  # noqa
from .client import *  # noqa
from .protocol import *  # noqa
from .ui_elems import *  # noqa
from .utils import *  # noqa
from .version import __version__  # noqa

# Don't pull in module objects like "client" when we do `from outleap import *`,
# only all of their exported vars.
__all__ = [
    *api_wrappers.__all__,
    *bridge.__all__,
    *client.__all__,
    *protocol.__all__,
    *utils.__all__,
    *ui_elems.__all__,
]



from __future__ import annotations

import abc
import pathlib
import uuid
from typing import *

from .client import COMMAND_PUMP, PUMP_NAME_TYPE, CommandPumpToken, LEAPClient
from .ui_elems import UI_PATH_TYPE, UIPath


class LEAPAPIWrapper(abc.ABC):
    """Base class for classes wrapping specific LEAP APIs"""

    PUMP_NAME: Optional[PUMP_NAME_TYPE] = None

    def __init__(self, client: LEAPClient, pump_name: Optional[PUMP_NAME_TYPE] = None):
        super().__init__()
        self._client = client
        self._pump_name = pump_name or self.PUMP_NAME
        assert self._pump_name


async def _data_unwrapper(data_fut: Awaitable[Dict], inner_elem: str) -> Any:
    """Unwraps part of the data future while allowing the request itself to remain synchronous"""
    # We want the request to be sent immediately, without requiring the request to be `await`ed first,
    # but that means that we have to return a `Coroutine` that will pull the value out of the dict
    # rather than directly returning the `Future`.
    return (await data_fut)[inner_elem]


class CommandAPI(LEAPAPIWrapper):
    PUMP_NAME = COMMAND_PUMP

    def get_apis(self) -> Awaitable[dict]:
        """
        Get a list of all available LLEventAPI instances

        Returns a dict of API name -> API details
        """
        return self._client.command(self._pump_name, "getAPIs")

    def get_api(self, api_name: str) -> Awaitable[dict]:
        """Get details about a specific LLEventAPI instance, including supported methos"""
        return self._client.command(self._pump_name, "getAPI", {"api": api_name})

    def ping(self) -> Awaitable[None]:
        """Send a ping and await the pong"""
        return self._client.command(self._pump_name, "ping")

    def start_listening(self, listener_name: str, source_pump: PUMP_NAME_TYPE) -> Awaitable[bool]:
        """Start listening on a specific pump, using `listener_name`"""
        if isinstance(source_pump, CommandPumpToken):
            source_pump = self._client.cmd_pump
        fut = self._client.command(
            self._pump_name,
            "listen",
            {
                "listener": listener_name,
                "source": source_pump,
            },
        )
        return _data_unwrapper(fut, "status")

    def stop_listening(self, listener_name: str, source_pump: PUMP_NAME_TYPE) -> Awaitable[bool]:
        """Stop `listener_name` from listening on a specific pump"""
        if isinstance(source_pump, CommandPumpToken):
            source_pump = self._client.cmd_pump
        fut = self._client.command(
            self._pump_name,
            "stoplistening",
            {
                "listener": listener_name,
                "source": source_pump,
            },
        )
        return _data_unwrapper(fut, "status")


class LLWindowAPI(LEAPAPIWrapper):
    PUMP_NAME = "LLWindow"

    MASK_TYPE = Optional[Collection[str]]
    KEYCODE_TYPE = Optional[int]
    KEYSYM_TYPE = Optional[str]
    CHAR_TYPE = Optional[str]
    MOUSE_COORD_TYPE: Optional[int]

    def _convert_key_payload(
        self,
        /,
        *,
        keycode: KEYCODE_TYPE,
        keysym: KEYSYM_TYPE,
        char: CHAR_TYPE,
        mask: MASK_TYPE,
        path: UI_PATH_TYPE,
    ) -> Dict:
        if keycode is not None:
            payload = {"keycode": keycode}
        elif keysym is not None:
            payload = {"keysym": keysym}
        elif char is not None:
            payload = {"char": char}
        else:
            raise ValueError("Didn't have one of keycode, keysym or char")

        if path:
            payload["path"] = str(path)
        if mask:
            payload["mask"] = mask

        return payload

    def key_down(
        self,
        /,
        *,
        mask: MASK_TYPE = None,
        keycode: KEYCODE_TYPE = None,
        keysym: KEYSYM_TYPE = None,
        char: CHAR_TYPE = None,
        path: UI_PATH_TYPE = None,
    ) -> None:
        """Simulate a key being pressed down"""
        payload = self._convert_key_payload(keysym=keysym, keycode=keycode, char=char, mask=mask, path=path)
        self._client.void_command(self._pump_name, "keyDown", payload)

    def key_up(
        self,
        /,
        *,
        mask: MASK_TYPE = None,
        keycode: KEYCODE_TYPE = None,
        keysym: KEYSYM_TYPE = None,
        char: CHAR_TYPE = None,
        path: UI_PATH_TYPE = None,
    ) -> None:
        """Simulate a key being released"""
        payload = self._convert_key_payload(keysym=keysym, keycode=keycode, char=char, mask=mask, path=path)
        self._client.void_command(self._pump_name, "keyUp", payload)

    def key_press(
        self,
        /,
        *,
        mask: MASK_TYPE = None,
        keycode: KEYCODE_TYPE = None,
        keysym: KEYSYM_TYPE = None,
        char: CHAR_TYPE = None,
        path: UI_PATH_TYPE = None,
    ) -> None:
        """Simulate a key being pressed down and immediately released"""
        self.key_down(mask=mask, keycode=keycode, keysym=keysym, char=char, path=path)
        self.key_up(mask=mask, keycode=keycode, keysym=keysym, char=char, path=path)

    def text_input(self, text_input: str, path: UI_PATH_TYPE = None) -> None:
        """Simulate a user typing a string of text"""
        # TODO: Uhhhhh I can't see how the key* APIs could possibly handle i18n correctly,
        #  what with all the U8s. Maybe I'm just dumb?
        for char in text_input:
            self.key_press(char=char, path=path)

    async def get_paths(self, under: Optional[UI_PATH_TYPE] = None) -> List[UIPath]:
        """Get all UI paths under the root, or under a path if specified"""
        if not under:
            under = ""
        resp = await self._client.command(self._pump_name, "getPaths", {"under": str(under)})
        if error := resp.get("error"):
            raise ValueError(error)
        return [UIPath(path) for path in resp.get("paths", [])]

    async def get_info(self, path: UI_PATH_TYPE) -> Dict:
        """Get info about an element specified by path"""
        return await self._client.command(self._pump_name, "getInfo", {"path": str(path)})

    def _build_mouse_payload(
        self,
        /,
        *,
        x: MOUSE_COORD_TYPE,
        y: MOUSE_COORD_TYPE,
        path: UI_PATH_TYPE,
        mask: MASK_TYPE,
        button: str = None,
    ) -> Dict:
        if path is not None:
            payload = {"path": str(path)}
        elif x is not None and y is not None:
            payload = {"x": x, "y": y}
        else:
            raise ValueError("Didn't have one of x + y or path")

        if mask:
            payload["mask"] = mask
        if button:
            payload["button"] = button

        return payload

    def mouse_down(
        self,
        /,
        *,
        x: MOUSE_COORD_TYPE = None,
        y: MOUSE_COORD_TYPE = None,
        path: UI_PATH_TYPE = None,
        mask: MASK_TYPE = None,
        button: str,
    ) -> Awaitable[Dict]:
        """Simulate a mouse down event occurring at a coordinate or UI element path"""
        payload = self._build_mouse_payload(x=x, y=y, path=path, mask=mask, button=button)
        return self._client.command(self._pump_name, "mouseDown", payload)

    def mouse_up(
        self,
        /,
        *,
        x: MOUSE_COORD_TYPE = None,
        y: MOUSE_COORD_TYPE = None,
        path: UI_PATH_TYPE = None,
        mask: MASK_TYPE = None,
        button: str,
    ) -> Awaitable[Dict]:
        """Simulate a mouse up event occurring at a coordinate or UI element path"""
        payload = self._build_mouse_payload(x=x, y=y, path=path, mask=mask, button=button)
        return self._client.command(self._pump_name, "mouseUp", payload)

    def mouse_click(
        self,
        /,
        *,
        x: MOUSE_COORD_TYPE = None,
        y: MOUSE_COORD_TYPE = None,
        path: UI_PATH_TYPE = None,
        mask: MASK_TYPE = None,
        button: str,
    ) -> Awaitable[Dict]:
        """Simulate a mouse down and immediately following mouse up event"""
        # We're going to ignore the mouseDown response, so use void_command instead.
        # Most side effects are actually executed on mouseUp.
        self._client.void_command(
            self._pump_name,
            "mouseDown",
            self._build_mouse_payload(x=x, y=y, path=path, mask=mask, button=button),
        )
        return self.mouse_up(x=x, y=y, path=path, mask=mask, button=button)

    def mouse_move(
        self, /, *, x: MOUSE_COORD_TYPE = None, y: MOUSE_COORD_TYPE = None, path: UI_PATH_TYPE = None
    ) -> Awaitable[Dict]:
        """Move the mouse to the coordinates or path specified"""
        payload = self._build_mouse_payload(x=x, y=y, path=path, mask=None)
        return self._client.command(self._pump_name, "mouseMove", payload)

    def mouse_scroll(self, clicks: int) -> None:
        """Act as if the scroll wheel has been moved `clicks` amount. May be negative"""
        self._client.command(self._pump_name, "mouseScroll", {"clicks": clicks})


class LLUIAPI(LEAPAPIWrapper):
    PUMP_NAME = "UI"

    def call(self, function: str, parameter: Any = None) -> None:
        """
        Invoke the `function` operation as if from a menu or button click, passing `parameter`

        Can call most things registered through `LLUICtrl::CommitCallbackRegistry`.
        """
        self._client.void_command(self._pump_name, "call", {"function": function, "parameter": parameter})

    def get_value(self, path: UI_PATH_TYPE) -> Awaitable[Any]:
        """For the UI control identified by `path`, return the current value in `value`"""
        resp_fut = self._client.command(self._pump_name, "getValue", {"path": str(path)})
        return _data_unwrapper(resp_fut, "value")


class LLCommandDispatcherAPI(LEAPAPIWrapper):
    PUMP_NAME = "LLCommandDispatcher"

    def dispatch(
        self,
        cmd: str,
        /,
        *,
        params: Optional[Collection[str]] = None,
        query: Optional[Dict[str, str]] = None,
        trusted: bool = True,
    ) -> None:
        """Execute a command registered as an LLCommandHandler"""
        return self._client.void_command(
            self._pump_name,
            "dispatch",
            {
                "cmd": cmd,
                "params": params or [],
                "query": query or {},
                "trusted": trusted,
            },
        )

    def enumerate(self) -> Awaitable[Dict]:
        """Get map of registered LLCommandHandlers, containing name, key, and (e.g.) untrusted flag"""
        return self._client.command(self._pump_name, "enumerate")


class LLViewerControlAPI(LEAPAPIWrapper):
    PUMP_NAME = "LLViewerControl"

    def get(self, group: str, key: str) -> Awaitable[Any]:
        """
        Get value of a Control (config) key

        `group` can be one of "CrashSettings", "Global", "PerAccount", "Warnings".
        """
        return self._client.command(self._pump_name, "get", {"key": key, "group": group})

    def vars(self, group: str) -> Awaitable[List[Dict]]:
        """Return a list of dicts of controls in `group`"""
        resp_fut = self._client.command(self._pump_name, "vars", {"group": group})
        return _data_unwrapper(resp_fut, "vars")

    def set(self, group: str, key: str, value: Any) -> None:
        """
        Set a configuration value

        TODO: error handling based on "error" field in resp?
        """
        self._client.void_command(self._pump_name, "set", {"key": key, "group": group, "value": value})


class LLViewerWindowAPI(LEAPAPIWrapper):
    PUMP_NAME = "LLViewerWindow"

    def save_snapshot(
        self,
        filename: Union[str, pathlib.Path],
        /,
        *,
        width: Optional[int] = None,  # Uses current dimensions if not specified
        height: Optional[int] = None,
        show_hud: bool = True,
        show_ui: bool = True,
        rebuild: bool = False,  # I guess rebuild the UI or not before snapshot
        snap_type: str = "COLOR",  # may be "DEPTH" or "COLOR"
    ) -> Awaitable[bool]:
        """Save a snapshot to the local disk"""
        extras = {}
        if width is not None:
            extras["width"] = width
        if height is not None:
            extras["height"] = height
        fut = self._client.command(
            self._pump_name,
            "saveSnapshot",
            {
                "filename": str(filename),
                "showhud": show_hud,
                "showui": show_ui,
                "rebuild": rebuild,
                "type": snap_type,
                **extras,
            },
        )
        return _data_unwrapper(fut, "ok")

    def request_reshape(self, width: int, height: int) -> None:
        """Request the window be resized to the specified dimensions"""
        self._client.void_command(self._pump_name, "requestReshape", {"w": width, "h": height})


class LLAgentAPI(LEAPAPIWrapper):
    PUMP_NAME = "LLAgent"

    def look_at(
        self,
        /,
        *,
        obj_uuid: Optional[uuid.UUID] = None,
        position: Optional[Sequence[float]] = None,
        lookat_type: int = 8,
    ):
        """Look at either a specific `obj_uuid` or the closest object to `position`"""
        payload = {"type": lookat_type}
        if obj_uuid:
            payload["obj_uuid"] = obj_uuid
        elif position:
            payload["position"] = list(position)
        else:
            raise ValueError("Must specify either obj_uuid or position")
        self._client.void_command(self._pump_name, "lookAt", payload)

    def get_auto_pilot(self) -> Awaitable[Dict]:
        """Get information about current state of the autopilot system"""
        return self._client.command(self._pump_name, "getAutoPilot", {})


class LLFloaterRegAPI(LEAPAPIWrapper):
    PUMP_NAME = "LLFloaterReg"

    def get_build_map(self) -> Awaitable[Dict]:
        """Get a map of floater names and their XUI xml files"""
        return self._client.command(self._pump_name, "getBuildMap", {})

    def show_instance(self, name: str, key: Any = None, focus: bool = False) -> None:
        """
        Show an instance of a floater

        `key` may contain specific data to bootstrap creating the floater, for
        example an item ID.
        """
        self._client.void_command(
            self._pump_name,
            "showInstance",
            {
                "name": name,
                "key": key,
                "focus": focus,
            },
        )

    def hide_instance(self, name: str, key: Any = None) -> None:
        """Hide an instance of a floater"""
        self._client.void_command(
            self._pump_name,
            "hideInstance",
            {
                "name": name,
                "key": key,
            },
        )

    def toggle_instance(self, name: str, key: Any = None) -> None:
        """Toggle visibility of an instance of a floater"""
        self._client.void_command(
            self._pump_name,
            "toggleInstance",
            {
                "name": name,
                "key": key,
            },
        )

    def is_instance_visible(self, name: str, key: Any = None) -> Awaitable[bool]:
        """Return whether an instance was visible"""
        fut = self._client.command(
            self._pump_name,
            "instanceVisible",
            {
                "name": name,
                "key": key,
            },
        )
        return _data_unwrapper(fut, "visible")

    def click_button(self, name: str, button: str, key: Any = None) -> Awaitable[Dict]:
        """Click a button on an instance of a floater, potentially returning failure details"""
        return self._client.command(
            self._pump_name,
            "clickButton",
            {
                "name": name,
                "key": key,
                "button": button,
            },
        )


class LLPuppetryAPI(LEAPAPIWrapper):
    PUMP_NAME = "puppetry"
    OP_KEY: str = "command"

    def get_camera(self) -> Awaitable[int]:
        """Request camera number: returns ["camera_id"]"""
        fut = self._client.command(self._pump_name, "get_camera", {}, op_key=self.OP_KEY)
        return _data_unwrapper(fut, "camera_id")

    def set_camera(self, camera_id: int) -> None:
        """Request camera number: returns ["camera_id"]"""
        payload = {"camera_id": camera_id}
        self._client.void_command(self._pump_name, "set_camera", payload, op_key=self.OP_KEY)

    def send_skeleton(self) -> None:
        """
        Request skeleton data

        Response will be sent over the "puppetry.command" listener as a "set_skeleton"
        """
        self._client.void_command(self._pump_name, "send_skeleton", {}, op_key=self.OP_KEY)

    def move(self, joint_data: Dict[str, Dict]) -> None:
        """
        Send puppet movement data

        Expected data format:
            {'joint_name':{'param_name':[r1.23,r4.56,r7.89]}, ...}
        Where:
            joint_name = e.g. mWristLeft
            param_name = rot | pos | scale | eff
            param value = array of 3 floats [x,y,z]
        """
        self._client.void_command(self._pump_name, "move", joint_data, op_key=self.OP_KEY)


__all__ = [
    "CommandAPI",
    "LLUIAPI",
    "LLAgentAPI",
    "LLWindowAPI",
    "LLViewerControlAPI",
    "LLViewerWindowAPI",
    "LLCommandDispatcherAPI",
    "LLFloaterRegAPI",
    "LLPuppetryAPI",
    "LEAPAPIWrapper",
]



import asyncio
import contextlib
import enum
import logging
import os
import signal
import sys
import tempfile
import weakref
from typing import *

import pkg_resources
from PySide6 import QtCore, QtGui, QtWidgets
from qasync import QEventLoop, asyncSlot

import outleap
from outleap import LEAPClient, UIElement, UIPath
from outleap.qt_helpers import GUIInteractionManager, loadUi

LOG = logging.getLogger(__name__)


def get_resource_filename(resource_filename: str):
    return pkg_resources.resource_filename("outleap", resource_filename)


class ElemTreeHeader(enum.IntEnum):
    Name = 0


MAIN_WINDOW_UI_PATH = get_resource_filename("scripts/ui/inspector.ui")


@contextlib.contextmanager
def temp_file_path():
    """Create a temporary file, yielding path and deleting it after"""
    # Windows NT can't have two open FHs on a tempfile, so we need
    # handle deletion ourselves :/
    with tempfile.NamedTemporaryFile(delete=False) as f:
        f.close()
        try:
            yield f.name
        finally:
            os.remove(f.name)


def _calc_clipped_rect(pix: QtGui.QPixmap, elem: outleap.UIElement) -> QtCore.QRect:
    base_rect = _elem_rect_to_qrect(pix, elem.rect)
    while elem := elem.parent:
        # Should also be clipped to the intersection of all parent rects,
        # some scrollers have offscreen rects that can't really be shown
        # in a screenshot.
        parent_rect = _elem_rect_to_qrect(pix, elem.rect)
        base_rect = base_rect.intersected(parent_rect)
    return base_rect


def _elem_rect_to_qrect(pix: QtGui.QPixmap, rect: outleap.UIRect) -> QtCore.QRect:
    # SL y origin is at the bottom, but Qt wants it at the top!
    # Need pixmap dimensions to convert.
    return QtCore.QRect(
        rect.left,
        pix.height() - rect.top,
        max(0, rect.right - rect.left),
        max(0, rect.top - rect.bottom),
    )


class LEAPInspectorGUI(QtWidgets.QMainWindow):
    lineEditFind: QtWidgets.QLineEdit
    treeElems: QtWidgets.QTreeWidget
    labelElemScreenshot: QtWidgets.QLabel
    textElemProperties: QtWidgets.QTextEdit
    btnRefresh: QtWidgets.QPushButton
    btnClickElem: QtWidgets.QPushButton
    btnSaveRendered: QtWidgets.QPushButton
    graphicsElemScreenshot: QtWidgets.QGraphicsView

    def __init__(self, client: outleap.LEAPClient, client_is_owned=False):
        super().__init__()

        loadUi(MAIN_WINDOW_UI_PATH, self)

        self.client = client
        self._client_is_owned = client_is_owned
        self._filter = ""

        self.interaction_manager = GUIInteractionManager(self)
        self.window_api = outleap.LLWindowAPI(client)
        self.viewer_window_api = outleap.LLViewerWindowAPI(self.client)
        self._element_tree = outleap.UIElementTree(self.window_api)
        self._items_by_path: Dict[UIPath, QtWidgets.QTreeWidgetItem] = weakref.WeakValueDictionary()  # noqa

        self.treeElems.setColumnCount(len(ElemTreeHeader))
        self.treeElems.setHeaderLabels(tuple(x.name for x in ElemTreeHeader))
        self.treeElems.header().setStretchLastSection(True)
        self.treeElems.selectionModel().selectionChanged.connect(self.updateSelectedElemInfo)
        self.btnRefresh.clicked.connect(self.reloadFromTree)
        self.lineEditFind.editingFinished.connect(self.filterChanged)

        self.btnClickElem.clicked.connect(self.clickElem)
        self.btnSaveRendered.clicked.connect(self.saveRendered)

        self.sceneElemScreenshot = QtWidgets.QGraphicsScene()
        self.graphicsElemScreenshot.setScene(self.sceneElemScreenshot)

    def showEvent(self, event: QtGui.QShowEvent) -> None:
        # Don't need to create_task() because this is an `asyncSlot()`
        self.reloadFromTree()
        return super().showEvent(event)

    def closeEvent(self, event: QtGui.QCloseEvent) -> None:
        if self._client_is_owned:
            try:
                self.client.disconnect()
            except Exception as e:
                logging.exception(e)
        return super().closeEvent(event)

    def _getSelectedPath(self) -> Optional[UIPath]:
        if selected := self.treeElems.selectedItems():
            return selected[0].data(ElemTreeHeader.Name, QtCore.Qt.UserRole)
        return None

    @asyncSlot()
    async def reloadFromTree(self):
        selected_path = self._getSelectedPath()

        self.treeElems.clear()
        self._items_by_path.clear()

        await self._element_tree.refresh()

        self.treeElems.setUpdatesEnabled(False)
        self._addChildren(self._element_tree.root_children, self.treeElems.invisibleRootItem())
        self.applyFilter()
        self.treeElems.setUpdatesEnabled(True)

        self.treeElems.sortByColumn(ElemTreeHeader.Name, QtCore.Qt.AscendingOrder)

        if selected_path and selected_path in self._items_by_path:
            item = self._items_by_path[selected_path]
            self.treeElems.scrollToItem(item)
            self.treeElems.setCurrentItem(item)

    def filterChanged(self):
        self._filter = self.lineEditFind.text()
        self.applyFilter()

    def applyFilter(self):
        filter_text = self._filter.lower()

        def _set_path_item_hidden(path: UIPath, hidden: bool):
            item = self._items_by_path.get(path)
            if item is not None:
                item.setHidden(hidden)
                # Only modify item expansion values if we actually specified
                # a filter. Don't contract everything if we clear out a filter,
                # so it's easy to see previously-filtered items in their full
                # context.
                if filter_text:
                    item.setExpanded(bool(filter_text and not hidden))

        def _unhide_ancestry(path: UIPath):
            _set_path_item_hidden(path, False)
            if path.parent != path:
                _unhide_ancestry(path.parent)

        self.treeElems.setUpdatesEnabled(False)
        for elem in self._element_tree:
            _set_path_item_hidden(elem, True)

        for path in self._element_tree:
            if filter_text in path.stem.lower():
                _unhide_ancestry(path)
        self.treeElems.setUpdatesEnabled(True)

        # Make it easier to find what was selected before the filter changed
        selected = self.treeElems.selectedItems()
        if selected and not selected[0].isHidden():
            self.treeElems.scrollToItem(selected[0])

    def _addChildren(self, nodes: Sequence[UIElement], parent: QtWidgets.QTreeWidgetItem):
        items = []
        # Batch up everything so that we can add it all to the tree at once.
        for node in nodes:
            # The type signature implies `None` is wrong here, but it isn't.
            item = QtWidgets.QTreeWidgetItem(None, [node.path.stem])  # noqa
            item.setData(ElemTreeHeader.Name, QtCore.Qt.UserRole, node.path)
            self._items_by_path[node.path] = item
            items.append(item)
            self._addChildren(node.children, item)
        parent.addChildren(items)

    @asyncSlot()
    async def updateSelectedElemInfo(self, *args):
        self.textElemProperties.setPlainText("")
        self.btnClickElem.setEnabled(False)
        self.btnSaveRendered.setEnabled(False)
        self.sceneElemScreenshot.clear()

        if not (selected_path := self._getSelectedPath()):
            return
        self.btnClickElem.setEnabled(True)

        # Display some info about the element this item refers to
        elem = self._element_tree[selected_path]
        # Refresh info for the element and its ancestors (we need updated rects)
        await asyncio.gather(*[e.refresh() for e in [elem, *elem.ancestors]])

        elem_str = ""
        for k, v in elem.to_dict().items():
            elem_str += f"{k}: {v}\n"
        self.textElemProperties.setPlainText(elem_str)
        print(elem_str, file=sys.stderr)

        # This is either an incomplete or invisible element, we can't show a preview.
        if not elem.info or not elem.visible_chain:
            return

        self.btnSaveRendered.setEnabled(True)

        # Draw the element preview
        # TODO: this is wasteful, cache screenshot pixmap for `n` seconds?
        pix_screenshot = QtGui.QPixmap()
        with temp_file_path() as path:
            await self.viewer_window_api.save_snapshot(path)
            pix_screenshot.load(path)

        # Clip scene and view to elem rect
        scene_rect = _calc_clipped_rect(pix_screenshot, elem)
        self.sceneElemScreenshot.addPixmap(pix_screenshot)
        self.sceneElemScreenshot.setSceneRect(scene_rect)
        self.graphicsElemScreenshot.fitInView(scene_rect, QtCore.Qt.KeepAspectRatio)

    @asyncSlot()
    async def clickElem(self):
        if not (selected_path := self._getSelectedPath()):
            return
        await self.window_api.mouse_click(path=selected_path, button="LEFT")
        # We clicked, which may have had an effect on the UI state.
        await self.updateSelectedElemInfo()

    @asyncSlot()
    async def saveRendered(self):
        if not (selected_path := self._getSelectedPath()):
            return
        elem = self._element_tree[selected_path]
        pix_screenshot = QtGui.QPixmap()
        with temp_file_path() as path:
            await self.viewer_window_api.save_snapshot(path)
            pix_screenshot.load(path)
        # Make a clipped copy of the screenshot
        clipped = pix_screenshot.copy(_calc_clipped_rect(pix_screenshot, elem))

        file_name = await self.interaction_manager.save_file(
            caption="Save Rendered Element", filter_str="PNG Images (*.png)", default_suffix="png"
        )
        if not file_name:
            return
        clipped.save(file_name, "PNG")


async def start_gui():
    signal.signal(signal.SIGINT, lambda *args: QtWidgets.QApplication.quit())
    client = await LEAPClient.create_stdio_client()
    window = LEAPInspectorGUI(client, client_is_owned=True)
    window.show()


def inspector_main():
    logging.basicConfig()
    QtCore.QCoreApplication.setAttribute(QtCore.Qt.AA_ShareOpenGLContexts)
    app = QtWidgets.QApplication(sys.argv)
    loop: asyncio.AbstractEventLoop = QEventLoop(app)
    asyncio.set_event_loop(loop)

    print(
        "Running in direct LEAP execution mode.\n"
        "If you're seeing this anywhere other than the viewer logs, "
        "you probably messed up, the viewer should be executing this!\n"
        "Try adding a '--tcp' argument!",
        file=sys.stderr,
    )

    loop.run_until_complete(start_gui())
    loop.run_forever()


if __name__ == "__main__":
    inspector_main()



"""
Stub for forwarding LEAP stdin/stdout to a LEAP receiver over TCP using netcat.

Really not much use to anyone but me until viewers correctly un-gate LEAP access :)
Hint: uncomment https://vcs.firestormviewer.org/phoenix-firestorm/files/cf85e854/indra/newview/llappviewer.cpp#L1398-1420

Usage: While an outleap TCP receiver is running
  ./firestorm --leap outleap-agent
"""
import asyncio
import multiprocessing

from outleap.utils import connect_stdin_stdout


async def _forward_stream(src_reader: asyncio.StreamReader, dst_writer: asyncio.StreamWriter):
    while not src_reader.at_eof() and not dst_writer.is_closing():
        dst_writer.write(await src_reader.read(0xFF00))
        await dst_writer.drain()


async def amain():
    serv_reader, serv_writer = await asyncio.open_connection("127.0.0.1", 9063, limit=10_000_000)
    stdio_reader, stdio_writer = await connect_stdin_stdout()

    try:
        agent_to_serv_fut = asyncio.create_task(_forward_stream(stdio_reader, serv_writer))
        serv_to_agent_fut = asyncio.create_task(_forward_stream(serv_reader, stdio_writer))
        await asyncio.gather(agent_to_serv_fut, serv_to_agent_fut)
    except asyncio.CancelledError:
        pass


def agent_main():
    loop = asyncio.get_event_loop_policy().get_event_loop()
    loop.run_until_complete(amain())


if __name__ == "__main__":
    multiprocessing.freeze_support()
    agent_main()



import asyncio
import html
from typing import *

from PySide6 import QtWidgets
from PySide6.QtCore import QMetaObject
from PySide6.QtUiTools import QUiLoader


class UiLoader(QUiLoader):
    """
    Subclass :class:`~PySide.QtUiTools.QUiLoader` to create the user interface
    in a base instance.

    Unlike :class:`~PySide.QtUiTools.QUiLoader` itself this class does not
    create a new instance of the top-level widget, but creates the user
    interface in an existing instance of the top-level class.

    This mimics the behaviour of :func:`PyQt4.uic.loadUi`.
    """

    def __init__(self, base_instance, custom_widgets=None):
        """
        Create a loader for the given ``baseinstance``.

        The user interface is created in ``baseinstance``, which must be an
        instance of the top-level class in the user interface to load, or a
        subclass thereof.

        ``customWidgets`` is a dictionary mapping from class name to class object
        for widgets that you've promoted in the Qt Designer interface. Usually,
        this should be done by calling registerCustomWidget on the QUiLoader, but
        with PySide 1.1.2 on Ubuntu 12.04 x86_64 this causes a segfault.

        ``parent`` is the parent object of this loader.
        """

        QUiLoader.__init__(self, base_instance)
        self.baseInstance = base_instance
        self.customWidgets = custom_widgets

    def createWidget(self, class_name, parent=None, name=""):
        """
        Function that is called for each widget defined in ui file,
        overridden here to populate baseinstance instead.
        """

        if parent is None and self.baseInstance:
            # supposed to create the top-level widget, return the base instance
            # instead
            return self.baseInstance

        else:
            if class_name in self.availableWidgets():
                # create a new widget for child widgets
                widget = QUiLoader.createWidget(self, class_name, parent, name)

            else:
                # if not in the list of availableWidgets, must be a custom widget
                # this will raise KeyError if the user has not supplied the
                # relevant class_name in the dictionary, or TypeError, if
                # customWidgets is None
                try:
                    widget = self.customWidgets[class_name](parent)

                except (TypeError, KeyError):
                    raise Exception(
                        f"No custom widget {class_name} found in customWidgets param of UiLoader __init__."
                    )

            if self.baseInstance:
                # set an attribute for the new child widget on the base
                # instance, just like PyQt4.uic.loadUi does.
                setattr(self.baseInstance, name, widget)

            return widget


def loadUi(uifile, baseinstance=None, custom_widgets=None, working_directory=None):
    """
    Dynamically load a user interface from the given ``uifile``.

    ``uifile`` is a string containing a file name of the UI file to load.

    If ``baseinstance`` is ``None``, a new instance of the top-level widget
    will be created.  Otherwise, the user interface is created within the given
    ``baseinstance``.  In this case ``baseinstance`` must be an instance of the
    top-level widget class in the UI file to load, or a subclass thereof.  In
    other words, if you've created a ``QMainWindow`` interface in the designer,
    ``baseinstance`` must be a ``QMainWindow`` or a subclass thereof, too.  You
    cannot load a ``QMainWindow`` UI file with a plain
    :class:`~PySide.QtGui.QWidget` as ``baseinstance``.

    ``customWidgets`` is a dictionary mapping from class name to class object
    for widgets that you've promoted in the Qt Designer interface. Usually,
    this should be done by calling registerCustomWidget on the QUiLoader, but
    with PySide 1.1.2 on Ubuntu 12.04 x86_64 this causes a segfault.

    :method:`~PySide.QtCore.QMetaObject.connectSlotsByName()` is called on the
    created user interface, so you can implemented your slots according to its
    conventions in your widget class.

    Return ``baseinstance``, if ``baseinstance`` is not ``None``.  Otherwise
    return the newly created instance of the user interface.
    """

    loader = UiLoader(baseinstance, custom_widgets)

    if working_directory is not None:
        loader.setWorkingDirectory(working_directory)

    widget = loader.load(uifile)
    QMetaObject.connectSlotsByName(widget)
    return widget


def show_error_message(error_msg, parent=None):
    error_dialog = QtWidgets.QErrorMessage(parent=parent)
    # No obvious way to set this to plaintext, yuck...
    error_dialog.showMessage(html.escape(error_msg))
    error_dialog.exec()
    error_dialog.raise_()


class GUIInteractionManager:
    def __init__(self, parent: QtWidgets.QWidget):
        self._parent = parent

    def _dialog_async_exec(self, dialog: QtWidgets.QDialog):
        future = asyncio.Future()
        dialog.finished.connect(lambda r: future.set_result(r))
        dialog.open()
        return future

    async def _file_dialog(
        self,
        caption: str,
        directory: str,
        filter_str: str,
        mode: QtWidgets.QFileDialog.FileMode,
        default_suffix: str = "",
    ) -> Tuple[bool, QtWidgets.QFileDialog]:
        dialog = QtWidgets.QFileDialog(self._parent, caption=caption, directory=directory, filter=filter_str)
        dialog.setFileMode(mode)
        if mode == QtWidgets.QFileDialog.FileMode.AnyFile:
            dialog.setAcceptMode(QtWidgets.QFileDialog.AcceptMode.AcceptSave)
        if default_suffix:
            dialog.setDefaultSuffix(default_suffix)
        res = await self._dialog_async_exec(dialog)
        return res, dialog

    async def open_files(self, caption: str = "", directory: str = "", filter_str: str = "") -> List[str]:
        res, dialog = await self._file_dialog(
            caption, directory, filter_str, QtWidgets.QFileDialog.FileMode.ExistingFiles
        )
        if not res:
            return []
        return dialog.selectedFiles()

    async def open_file(self, caption: str = "", directory: str = "", filter_str: str = "") -> Optional[str]:
        res, dialog = await self._file_dialog(
            caption, directory, filter_str, QtWidgets.QFileDialog.FileMode.ExistingFile
        )
        if not res:
            return None
        return dialog.selectedFiles()[0]

    async def open_dir(self, caption: str = "", directory: str = "", filter_str: str = "") -> Optional[str]:
        res, dialog = await self._file_dialog(
            caption, directory, filter_str, QtWidgets.QFileDialog.FileMode.Directory
        )
        if not res:
            return None
        return dialog.selectedFiles()[0]

    async def save_file(
        self, caption: str = "", directory: str = "", filter_str: str = "", default_suffix: str = ""
    ) -> Optional[str]:
        res, dialog = await self._file_dialog(
            caption,
            directory,
            filter_str,
            QtWidgets.QFileDialog.FileMode.AnyFile,
            default_suffix,
        )
        if not res or not dialog.selectedFiles():
            return None
        return dialog.selectedFiles()[0]

    async def confirm(self, title: str, caption: str) -> bool:
        msg = QtWidgets.QMessageBox(
            QtWidgets.QMessageBox.Icon.Question,
            title,
            caption,
            QtWidgets.QMessageBox.Ok | QtWidgets.QMessageBox.Cancel,
            self._parent,
        )
        fut = asyncio.Future()
        msg.finished.connect(lambda r: fut.set_result(r))
        msg.open()
        return (await fut) == QtWidgets.QMessageBox.Ok



import outleap

from . import BaseClientTest


class TestWrappers(BaseClientTest):
    async def test_mouse_click(self):
        self._write_welcome()
        await self.client.connect()
        llwindow_api = outleap.LLWindowAPI(self.client)

        # Make sure the request looks like what we'd expect
        fut = llwindow_api.mouse_click(x=0, y=1, mask=["CTL"], button="LEFT")
        self.assertDictEqual(
            {
                "pump": "LLWindow",
                "data": {
                    "x": 0,
                    "y": 1,
                    "mask": ["CTL"],
                    "button": "LEFT",
                    "op": "mouseUp",
                    "reply": "reply_pump",
                    "reqid": 1,
                },
            },
            self.protocol.sent_messages[-1],
        )
        self.assertEqual(2, len(self.protocol.sent_messages[-1]))

        # Make sure the response is handled correctly
        self.protocol.inbound_messages.put_nowait(
            {
                "pump": "reply_pump",
                "data": {
                    "handled": True,
                    "reqid": 1,
                },
            }
        )
        await fut

    async def test_key_press(self):
        self._write_welcome()
        await self.client.connect()
        llwindow_api = outleap.LLWindowAPI(self.client)

        # Make sure the request looks like what we'd expect
        llwindow_api.key_press(char="f", mask=["CTL"])
        self.assertDictEqual(
            {"pump": "LLWindow", "data": {"char": "f", "mask": ["CTL"], "op": "keyUp"}},
            self.protocol.sent_messages[-1],
        )
        self.assertEqual(2, len(self.protocol.sent_messages[-1]))

    async def test_uipath_key_down(self):
        self._write_welcome()
        await self.client.connect()

        llwindow_api = outleap.LLWindowAPI(self.client)
        llwindow_api.key_down(char="f", path=outleap.UIPath.for_floater("foo") / "bar_elem")
        self.assertDictEqual(
            {
                "pump": "LLWindow",
                "data": {
                    "char": "f",
                    "path": "/main_view/menu_stack/world_panel/Floater View/foo/bar_elem",
                    "op": "keyDown",
                },
            },
            self.protocol.sent_messages[-1],
        )

    async def test_command_wrapper(self):
        self._write_welcome()
        await self.client.connect()

        llwindow_api = outleap.CommandAPI(self.client)
        fut = llwindow_api.get_api("foo")
        self.assertDictEqual(
            {
                "pump": self.client.cmd_pump,
                "data": {
                    "api": "foo",
                    "op": "getAPI",
                    "reply": "reply_pump",
                    "reqid": 1,
                },
            },
            self.protocol.sent_messages[-1],
        )
        self.protocol.inbound_messages.put_nowait(
            {
                "pump": "reply_pump",
                "data": {
                    "whatever": "foo",
                    "reqid": 1,
                },
            }
        )
        self.assertEqual("foo", (await fut)["whatever"])



#!/usr/bin/env python
"""
Simple example uses of the various LEAP APIs.

May be run either directly by the viewer, or through a TCP connect-back using --tcp.
"""

import asyncio
import logging
import pprint
import sys
from typing import *

from outleap import (
    COMMAND_PUMP,
    LLUIAPI,
    CommandAPI,
    LEAPBridgeServer,
    LEAPClient,
    LLCommandDispatcherAPI,
    LLFloaterRegAPI,
    LLViewerControlAPI,
    LLWindowAPI,
    UIPath,
)


async def client_connected(client: LEAPClient):
    printer = pprint.PrettyPrinter(stream=sys.stderr)
    # Kick off a request to get ops for each API supported by the viewer
    # Won't wait for a response from the viewer between each send
    api_futs: Dict[Awaitable, str] = {}
    # COMMAND_PUMP is a special pump and refers to whatever command pump was assigned
    # to us as part of the LEAP handshake. It's different every time.
    for api_name in (await client.command(COMMAND_PUMP, "getAPIs")).keys():
        api_fut = client.command(COMMAND_PUMP, "getAPI", {"api": api_name})
        api_futs[api_fut] = api_name

    # Wait for all of our getAPI commands to complete in parallel
    for fut in (await asyncio.wait(api_futs.keys()))[0]:
        # Print out which API this future even relates to
        print("=" * 5, api_futs[fut], "=" * 5, file=sys.stderr)
        # List supported ops for this api
        printer.pprint(await fut)

    # Subscribe to StartupState events within this scope
    async with client.listen_scoped("StartupState") as listener:
        # Get a single StartupState event then continue
        printer.pprint(await listener.get())

    # More manual version of above that gives you a Queue you can pass around
    # A None gets posted to the mainloop every time the viewer restarts the main loop,
    # so we can rely on _something_ being published to this.
    mainloop_listener = await client.listen("mainloop")
    try:
        printer.pprint(await mainloop_listener.get())
    finally:
        await client.stop_listening(mainloop_listener)

    # A simple command with a reply
    # Let's not use the fancy API wrappers for now.
    printer.pprint(await client.command("LLFloaterReg", "getBuildMap"))

    # A simple command that has no reply, or has a reply we don't care about.
    client.void_command("LLFloaterReg", "showInstance", {"name": "preferences"})

    # Simple connectivity check, this is largely the same as doing
    # `await client.command(COMMAND_PUMP, "ping")`, but with intellisense support.
    command_api = CommandAPI(client)
    await command_api.ping()

    # Print out all the commands supported by LLCommandDispatcher
    cmd_dispatcher_api = LLCommandDispatcherAPI(client)
    printer.pprint(await cmd_dispatcher_api.enumerate())

    # Spawn the test textbox floater, we can use the nice API wrapper for that.
    floater_reg_api = LLFloaterRegAPI(client)
    floater_reg_api.show_instance("test_textbox")

    # LEAP allows addressing UI elements by "path". We expose that through a pathlib-like interface
    # to allow composing UI element paths.
    textbox_path = UIPath.for_floater("floater_test_textbox") / "long_text_editor"
    # Click the "long_text_editor" in the test textbox floater.
    window_api = LLWindowAPI(client)
    await window_api.mouse_click(button="LEFT", path=textbox_path)

    # Clear out the textbox, note that this does _not_ work when path is specified!
    # TODO: clearing a textbox isn't so nice. CTL+A doesn't work as expected even without a path,
    #  it leaves a capital "A" in the text editor. We get rid of it by doing backspace right after.
    window_api.key_press(mask=["CTL"], keysym="a")
    window_api.key_press(keysym="Backsp")

    # Type some text
    window_api.text_input("Also I can type in here pretty good.")

    # Print out the value of the textbox we just typed in
    ui_api = LLUIAPI(client)
    printer.pprint(await ui_api.get_value(textbox_path))

    # But you don't need to explicitly give input focus like above, you can send keypresses
    # directly to a path.
    monospace_path = UIPath.for_floater("floater_test_textbox") / "monospace_text_editor"
    window_api.text_input("I typed in here by path.", path=monospace_path)

    # We can also access the viewer config to reason about viewer state.
    viewer_control_api = LLViewerControlAPI(client)
    printer.pprint(await viewer_control_api.get("Global", "StatsPilotFile"))
    # Print the first ten vars in the "Global" group
    printer.pprint((await viewer_control_api.vars("Global"))[:10])

    # Done, bye!
    client.disconnect()


def receiver_main():
    logging.basicConfig(level=logging.INFO)
    loop = asyncio.get_event_loop_policy().get_event_loop()

    args = sys.argv[1:]
    if args and args[0] == "--tcp":
        # In this mode viewers connect to the script over TCP,
        # and multiple viewers may be controlled by the same script at once.
        print("Ok, running in TCP daemon mode, will wait for connect-backs!", file=sys.stderr)
        server = LEAPBridgeServer(client_connected)
        coro = asyncio.start_server(server.handle_connection, "127.0.0.1", 9063)
        loop.run_until_complete(coro)
        loop.run_forever()
    else:
        # In this mode viewers directly execute the script via their --leap argument.
        # Only one viewer may be controlled.
        print(
            "Running in direct LEAP execution mode.\n"
            "If you're seeing this anywhere other than the viewer logs, "
            "you probably messed up, the viewer should be executing this!\n"
            "Try adding a '--tcp' argument!",
            file=sys.stderr,
        )

        async def _wrapper():
            async with await LEAPClient.create_stdio_client() as client:
                await client_connected(client)

        loop.run_until_complete(_wrapper())


if __name__ == "__main__":
    receiver_main()



import asyncio
import io
import os
import stat
import sys
import threading
import time
from typing import *


class HackySTDIOTransport(asyncio.Transport):
    def __init__(self):
        super().__init__()
        self.closed = False
        self.wrote_eof = False

    def can_write_eof(self) -> bool:
        return not self.closed and not self.wrote_eof

    def write_eof(self) -> None:
        assert self.can_write_eof()
        self.wrote_eof = True

    def is_closing(self) -> bool:
        return self.closed

    def is_reading(self) -> bool:
        return False

    def close(self) -> None:
        self.closed = True

    def write(self, data: Any) -> None:
        sys.stdout.buffer.write(data)
        sys.stdout.buffer.flush()


class HackySTDIOProtocol(asyncio.Protocol):
    def __init__(self):
        super().__init__()

    def data_received(self, data: bytes) -> None:
        pass

    def eof_received(self) -> Union[bool, None]:
        return False

    def resume_writing(self) -> None:
        pass

    def pause_writing(self) -> None:
        pass

    async def _drain_helper(self):
        # no-op, we flush immediately.
        pass


_READER_BUFFER_LIMIT = 10_000_000


def _stdin_feeder(
    transport: asyncio.Transport, reader: asyncio.StreamReader, loop: asyncio.AbstractEventLoop
):
    while not transport.is_closing():
        data = os.read(0, _READER_BUFFER_LIMIT)
        if not data:
            time.sleep(0.0000001)
            continue
        loop.call_soon_threadsafe(reader.feed_data, data)
    reader.feed_eof()


async def _make_hacky_threaded_stdio_rw() -> Tuple[asyncio.StreamReader, asyncio.StreamWriter]:
    # NB: Windows has no non-blocking IO for stdin and stdout, so we have to resort to weird
    # hacks. Do blocking IO in a separate thread and feed it to a StreamReader so that we can
    # keep API symmetry across stdio & TCP LEAP code. Otherwise, we end up with an
    # un-cancellable Future like in the run_in_executor() case. We don't care if the thread
    # is still blocking trying to read when we exit, only one thing is trying to read from
    # stdin!
    # See https://stackoverflow.com/questions/31510190/aysncio-cannot-read-stdin-on-windows
    #
    # TODO: Currently we also use this if we're reading from a non-pipe file descriptor on POSIX
    #  platforms, but that's probably unnecessary.
    loop = asyncio.get_event_loop()
    reader = asyncio.StreamReader(limit=_READER_BUFFER_LIMIT)
    protocol = HackySTDIOProtocol()
    transport = HackySTDIOTransport()
    writer = asyncio.StreamWriter(transport, protocol, reader, loop)

    # Runs forever and dies with the process.
    t = threading.Thread(target=_stdin_feeder, args=(transport, reader, loop), daemon=True)
    t.daemon = True
    t.start()
    return reader, writer


def _is_unix_pipe(fileno: int) -> bool:
    mode = os.fstat(fileno).st_mode
    return stat.S_ISFIFO(mode) or stat.S_ISSOCK(mode) or stat.S_ISCHR(mode)


# Modern Linux allows pipe sizes to be increased up to 1MB without special perms by default
# https://man7.org/linux/man-pages/man2/fcntl.2.html
_PIPE_BUFFER_SIZE = 1_000_000


def _patch_stdio_buffering():
    """
    Improve LEAP performance by giving stdin and stdout larger buffers

    Due to how the viewer's LEAP bridge is implemented, it will yield to the main loop
    if the child's stdin pipe fills. The default pipe size on Linux is about 8k. This
    leads to the LEAP write yielding for 0.05~s every time it writes an 8k chunk, which
    obviously leads to very bad perf for larger payloads:

    https://bitbucket.org/lindenlab/viewer/src/f83289d3a7e80bebe47f696f96aee1b7e64d1d69/indra/llcommon/llprocess.cpp#lines-228:235

    By increasing the size of the stdin / stdout pipes, we reduce the number of times
    the viewer must yield to the main loop while sending us the payload, reducing
    apparent function call time by 10x for larger payloads.

    Only works on Linux on Python 3.10+ due to relying on pipe implementation details
    """
    try:
        import fcntl
    except ImportError:
        # Will probably happen on OS X
        return
    try:
        # https://man7.org/linux/man-pages/man2/fcntl.2.html
        set_pipe_size = getattr(fcntl, "F_SETPIPE_SZ")
    except AttributeError:
        # Will probably happen on Python before 3.10
        return
    # Might fail, don't care. Best efforts only.
    pipes = [sys.stdin, sys.stdout]
    for pipe in pipes:
        try:
            if not _is_unix_pipe(pipe.fileno()):
                continue
            fcntl.fcntl(pipe.fileno(), set_pipe_size, _PIPE_BUFFER_SIZE)
        except (OSError, io.UnsupportedOperation):
            pass


async def connect_stdin_stdout() -> Tuple[asyncio.StreamReader, asyncio.StreamWriter]:
    """
    Get a StreamReader and StreamWriter for stdin and stdout, respectively
    """
    _patch_stdio_buffering()
    # Windows (and probably cygwin) need special logic for non-blocking use of STDIO.
    need_stdin_hack = any(sys.platform.startswith(x) for x in ("cygwin", "win32"))
    if not need_stdin_hack:
        # We also need it if stdin isn't actually a pipe, the pipe will fail to connect!
        if not _is_unix_pipe(sys.stdin.fileno()) or not _is_unix_pipe(sys.stdout.fileno()):
            need_stdin_hack = True
    if need_stdin_hack:
        return await _make_hacky_threaded_stdio_rw()
    loop = asyncio.get_event_loop()
    reader = asyncio.StreamReader(limit=_READER_BUFFER_LIMIT)
    protocol = asyncio.StreamReaderProtocol(reader)
    await loop.connect_read_pipe(lambda: protocol, sys.stdin)
    w_transport, w_protocol = await loop.connect_write_pipe(asyncio.streams.FlowControlMixin, sys.stdout)
    writer = asyncio.StreamWriter(w_transport, w_protocol, reader, loop)
    return reader, writer


__all__ = [
    "connect_stdin_stdout",
]
